import java.sql.*;
import java.util.*;


/**
 * A Database manipulation class :
 * - manage connection to DBS
 * - manage extraction of data and metadata
 */
public class Database {

	private String dsn;
	private String driver;
	private String user;
	private String pass;
	private String schemaName;

	private Connection conn;
	private ResultSetMetaData rsetmd;
	private DatabaseMetaData dbmd = null;
	private String catalog;
	private String types[];
	private Statement stmt = null;
	private Config connectionSettings;

	/**
	 * Default constructor
	 */
	public Database() {
		connectionSettings = new Config("connection.properties");
		this.driver = connectionSettings.getProperty("driver");
		this.dsn = connectionSettings.getProperty("dsn");
		this.user = connectionSettings.getProperty("user");
		this.pass = connectionSettings.getProperty("pass");
		this.schemaName = connectionSettings.getProperty("schemaName");
		types = new String[1];
		types[0] = "TABLES";
	}
	/**
	 * Open a database connection
	 */
	public void Connect() {
		if (conn != null)
			return;
		try {
			System.err.println("Opening database");
			Class.forName(driver).newInstance();
			conn = DriverManager.getConnection(dsn, user, pass);
			dbmd = conn.getMetaData();
		} catch (ClassNotFoundException ex) {
			System.err.println("[-]Cannot find the database driver classes.");
			System.exit(-1);
		} catch (SQLException ex) {
			System.err.println("[-]Cannot connect to this database.");
			System.exit(-1);
		} catch (Exception ex) {
			System.err.println(ex);
			System.exit(-1);
		}
	}				
	
	public Connection getConnection() {
		if (conn == null)
			Connect();
		return (conn);
	}
	/**
	 * Close the database connection
	 */
	public void Close() {
		try {
			if (conn != null) {
				conn.close();
				System.err.println("Closing database");
			}
		} catch (java.sql.SQLException e) {
			System.err.println("[-]Database.closeConnection() - SQLException: " + e);
		}
	}
	/**
	 * Retrieve the name of each table associated to a catalog and a schema
	 * @return the tablenames
	 */
	public String[] getTableNames() {		
		Vector tname = new Vector();
		try {
			ResultSet rs = dbmd.getTables(catalog, schemaName, null, null);
			while (rs.next())
				tname.addElement(rs.getString("TABLE_NAME"));
			rs.close();
			String[] tbnames = new String[tname.size()];
			for (int i = 0; i < tname.size(); i++)
				tbnames[i] = (String) tname.elementAt(i);
			return tbnames;
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
		return(null);
	}

	/**
	 * Retrieve the name of each column associated to a table
	 * @param table name
	 * @return the column names
	 */
	public String[] getColumnNames(String table) {		
		Vector tname = new Vector();
		try {
			ResultSet rs = dbmd.getColumns(catalog, schemaName, table, null);
			while (rs.next())
				tname.addElement(rs.getString("COLUMN_NAME"));
			rs.close();
			String[] tbnames = new String[tname.size()];
			for (int i = 0; i < tname.size(); i++)
				tbnames[i] = (String) tname.elementAt(i);
			return tbnames;
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
		return(null);
	}

	/**
	 * return the SchemaName associated to this DatabaseConnection
	 * @return the shchema's name
	 */
	public String getSchemaName() {
		return this.schemaName;
	}
	/**
	 * return the type of a given column
	 * @param table
	 * @param columnName
	 * @return type
	 */
	public String getColumnType(String table, String columnName) {
		Config typeMap = new Config("TypeMapping.properties");		
		try {
			if (table.length() > 0 && columnName.length() >0)
			{	
				DatabaseAnswer answer = Query("Select " + columnName + " from " + table);
				String columnType = answer.getResult().getMetaData().getColumnTypeName(1);										
				return  typeMap.getProperty(columnType);				
			}						
		} catch (Exception e) {
			System.out.println("Column type error" + table + "."+ columnName + e.getMessage());
		}
		return(null);
	}
	/**
	 * return the length(size) of a column
	 * 
	 * @param table name
	 * @param column name
	 * @return length
	 */
	public String getColumnLength(String table, String columnName) {
		Config typeMap = new Config("TypeMapping.properties");
		String columnLength = null;
		try {
			if (table.length() > 0 && columnName.length() > 0) {
				DatabaseAnswer answer = Query("Select " + columnName + " from " + table);				
				String columnType = answer.getResult().getMetaData().getColumnTypeName(1);
				if (typeMap.getProperty(columnType + "_LENGTH").equals("YES")) 
				{
					columnLength = Integer.toString(answer.getResult().getMetaData().getPrecision(1));										
				}										
				answer.close();
				return columnLength;
			}
		} catch (Exception e) {
			System.out.println("Column length error" + table + "." + columnName + e.getMessage());
		}
		return columnLength;
	}
	/**
	 * return the scale of a given column
	 * 
	 * @param table
	 * @param columnName
	 * @return scale
	 */
	public String getColumnScale(String table, String columnName) {
		Config typeMap = new Config("TypeMapping.properties");
		String columnScale = null;
		try {
			if (table.length() > 0 && columnName.length() > 0) {
				DatabaseAnswer answer = Query("Select " + columnName + " from " + table);
				String columnType = answer.getResult().getMetaData().getColumnTypeName(1);
				if (typeMap.getProperty(columnType + "_SCALE").equals("YES")) {
					columnScale = Integer.toString(answer.getResult().getMetaData().getScale(1));
				}
				answer.close();
				return columnScale;
			}
		} catch (Exception e) {
			System.out.println("Column scale error" + table + "." + columnName + e.getMessage());
		}
		return columnScale;
	}	
	
	
	/**
	 * return the primary keys list of a given table
	 * @param table
	 * @return
	 */
	public String[] getPrimaryKeys(String table) {		
		Vector pname = new Vector();
		try {
			ResultSet rs = dbmd.getPrimaryKeys(catalog, schemaName, table);
			while (rs.next())
				pname.addElement(rs.getString("COLUMN_NAME"));
			rs.close();
			String[] pknames = new String[pname.size()];
			for (int i = 0; i < pname.size(); i++)
				pknames[i] = (String) pname.elementAt(i);
			return pknames;
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
		return(null);
	}
	/**
	 * Retrieves a list of foreign key columns for a given table.
	 * 
	 * @param tableName =Table from which to retrieve FK information.
	 * @return A list of foreign keys [foreign column name,local column name,local table name] 
	 * @throws SQLException
	 */
	public String[][] getForeignKeys(String tableName) throws SQLException {		
		Vector fkname = new Vector();
		Vector fktname = new Vector();
		Vector fname = new Vector();
		try {			
			ResultSet rs = dbmd.getImportedKeys(catalog, schemaName, tableName);
			while (rs.next()) {
				fname.addElement(rs.getString("FKCOLUMN_NAME"));
				fkname.addElement(rs.getString("PKCOLUMN_NAME"));
				fktname.addElement(rs.getString("PKTABLE_NAME"));				
			}
			rs.close();
			String[][] fknames = new String[fkname.size()][3];	
			for (int i = 0; i < fkname.size(); i++) {
				fknames[i][0] = (String) fname.elementAt(i);
				fknames[i][1] = (String) fktname.elementAt(i);
				fknames[i][2] = (String) fkname.elementAt(i);
			}
			return fknames;
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
		return(null);
	}

	/**
	 * retrieve all the data value specific to a column (of a table)
	 * 
	 * @param table
	 * @param columnName
	 * @return the values of the table+column selected
	 */
	public String[] getColumnValues(String table, String columnName) {
		Vector tvalues = new Vector();
		try {
			DatabaseAnswer answer = Query("Select " + columnName + " from " + table);			
			while (answer.getResult().next())
				tvalues.addElement(answer.getResult().getString(columnName));
			answer.close();	// do not forget to close the answer, other case -> statement heap is exploding !
			String[] tbvalues = new String[tvalues.size()];
			for (int i = 0; i < tvalues.size(); i++)
				tbvalues[i] = (String) tvalues.elementAt(i);
			return tbvalues;
		} catch (Exception e) {
			System.out.println("getColumnValues : ");
			e.printStackTrace();
		}
		return(null);
	}
	/**
	 * execute an SQL query on this database
	 * @param sql = SQL query
	 * @return the resulset associated to the query
	 */
	public DatabaseAnswer Query(String sql) {
		try {
			Connection cnt = getConnection();
			if (cnt == null)
				throw new Exception("cannot connect to database.");
			Statement st = cnt.createStatement();
			ResultSet rs = st.executeQuery(sql);
			return (new DatabaseAnswer(st, rs));
		} catch (java.sql.SQLException e) {
			System.err.println("FATAL: SQLException " + e);
			System.exit(-1);
		} catch (java.lang.Exception e) {
			System.err.println("FATAL: Exception " + e);
			System.exit(-1);
		}
		return (null);
	}
	/**
	 * execute an SQL create/update/... code
	 * 
	 * @param sql = SQL action (in data definition langage) 
	 */
	public void Execute(String sql) {
		try {
			Statement stmt = conn.createStatement();
			stmt.execute(sql);
		} catch (Exception e) {
			System.out.println("update error: " + e.getMessage());
		}
	}
}
